/*
 * Copyright 2009 Mozilla Corporation.
 *
 * This is part of HarfBuzz, an OpenType Layout engine library.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Created by Jonathan Kew on 17/06/2009.
 */

#include "hb-cmap.h"
#include "hb-private.h"

struct _cmap_encoding_subtable_header
{
  guint16 platform_id;
  guint16 encoding_id;
  guint32 offset;
};
typedef struct _cmap_encoding_subtable_header cmap_encoding_subtable_header;

struct _cmap_table
{
  guint16   version;
  guint16   number_subtables;
  cmap_encoding_subtable_header encoding_subtable_headers[1];
};
typedef struct _cmap_table cmap_table;

struct _format_4_cmap
{
  guint16 format;
  guint16 length;
  guint16 language;
  guint16 seg_count_x_2;
  guint16 search_range;
  guint16 entry_selector;
  guint16 range_shift;
  
  guint16 arrays[1];
};
typedef struct _format_4_cmap format_4_cmap;

struct _format_12_group
{
  guint32 start_char_code;
  guint32 end_char_code;
  guint32 start_glyph_id;
};
typedef struct _format_12_group format_12_group;

struct _format_12_cmap
{
  guint16 format;
  guint16 reserved;
  guint32 length;
  guint32 language;
  guint32 num_groups;
  
  format_12_group groups[1];
};
typedef struct _format_12_cmap format_12_cmap;

/* TODO: format 14 support */

static hb_codepoint_t
_hb_cmap_subtable_format_4_char_to_glyph(const format_4_cmap *cmap4, hb_codepoint_t unicode_scalar)
{
  guint16 segCount;
  const guint16 *endCodes;
  const guint16 *startCodes;
  const guint16 *idDelta;
  const guint16 *idRangeOffset;
  guint16 probe;
  guint16 rangeShiftOver2;
  guint16 index;
  
  if (unicode_scalar >= 0x10000) {
    return 0;
  }
  
  segCount = hb_be_uint16_t(cmap4->seg_count_x_2) / 2;
  
  endCodes = &cmap4->arrays[0];
  startCodes = &cmap4->arrays[segCount + 1]; // +1 because of reserved word between arrays
  idDelta = &startCodes[segCount];
  idRangeOffset = &idDelta[segCount];
  
  probe = 1 << hb_be_uint16_t(cmap4->entry_selector);
  rangeShiftOver2 = hb_be_uint16_t(cmap4->range_shift) / 2;
  
  if (hb_be_uint16_t(startCodes[rangeShiftOver2]) <= unicode_scalar) {
    index = rangeShiftOver2;
  } else {
    index = 0;
  }
  
  while (probe > 1) {
    probe >>= 1;
    if (hb_be_uint16_t(startCodes[index + probe]) <= unicode_scalar) {
      index += probe;
    }
  }
  
  if (unicode_scalar >= hb_be_uint16_t(startCodes[index]) &&
      unicode_scalar <= hb_be_uint16_t(endCodes[index]))
  {
    guint16 result;
    if (idRangeOffset[index] == 0) {
      result = unicode_scalar;
    } else {
      guint16 offset = unicode_scalar - hb_be_uint16_t(startCodes[index]);
      const guint16 *glyphIndexTable = (const guint16*)((const char*)&idRangeOffset[index] +
							hb_be_uint16_t(idRangeOffset[index]));
      result = hb_be_uint16_t(glyphIndexTable[offset]);
    }
    
    /* note that this is unsigned 16-bit arithmetic, and may wrap around */
    result += hb_be_uint16_t(idDelta[index]);
    return result;
  }
  
  return 0;
}

/*
 * Return just the highest bit of the given value, i.e., the highest power of 2 that is <= value,
 * or zero if the input value is zero.
 */
static guint32
_find_highest_bit(guint32 value)
{
  /* propagate highest bit into all lower bits of the value */
  value |= (value >> 1);
  value |= (value >> 2);
  value |= (value >> 4);
  value |= (value >> 8);
  value |= (value >> 16);
  /* isolate the leftmost bit */
  return (value & ~(value >> 1));
}

static hb_codepoint_t
_hb_cmap_subtable_format_12_char_to_glyph(const format_12_cmap *cmap12, hb_codepoint_t unicode_scalar)
{
  guint32 numGroups = hb_be_uint32_t(cmap12->num_groups);
  guint32 powerOf2 = _find_highest_bit(numGroups);
  guint32 rangeOffset = numGroups - powerOf2;
  guint32 range = 0;
  guint32 startCharCode;
  
  const format_12_group *groups = &cmap12->groups[0];
  if (hb_be_uint32_t(groups[rangeOffset].start_char_code) <= unicode_scalar) {
    range = rangeOffset;
  }
  
  while (powerOf2 > 1) {
    powerOf2 >>= 1;
    if (hb_be_uint32_t(groups[range + powerOf2].start_char_code) <= unicode_scalar) {
      range += powerOf2;
    }
  }
  
  startCharCode = hb_be_uint32_t(groups[range].start_char_code);
  if (startCharCode <= unicode_scalar && hb_be_uint32_t(groups[range].end_char_code) >= unicode_scalar) {
    return hb_be_uint32_t(groups[range].start_glyph_id) + unicode_scalar - startCharCode;
  }
  
  return 0;
}

/*
 * Map unicode value to glyph ID through a given cmap subtable.
 * Currently supports only formats 4 and 12.
 */
hb_codepoint_t
hb_cmap_subtable_char_to_glyph(const char *cmap_subtable, hb_codepoint_t unicode_scalar)
{
  // format field is first 16-bit word of subtable
  guint16 cmap_format = hb_be_uint16_t(*(const guint16*)cmap_subtable);
  
  switch (cmap_format) {
    case 4:
      return _hb_cmap_subtable_format_4_char_to_glyph((const format_4_cmap*)cmap_subtable, unicode_scalar);
      
    case 12:
      return _hb_cmap_subtable_format_12_char_to_glyph((const format_12_cmap*)cmap_subtable, unicode_scalar);
  }
  
  return 0;
}

/*
 * Find our preferred subtable, returning a pointer to the data, and optionally its length
 * (in case the client wants to cache the subtable rather than hold onto the entire cmap).
 * Returns subtable_length = 0 if it cannot determine the length; in this case, the client
 * must assume that it extends to the end of the cmap table.
 */
const char*
hb_cmap_find_preferred_subtable(const char *cmap_data, guint32 *subtable_length)
{
  const cmap_table *cmap = (const cmap_table*)cmap_data;
  guint16 index;
  guint16 nSubtables = hb_be_uint16_t(cmap->number_subtables);
  const guint16 *subtable;
  guint32 msOffset1 = 0, msOffset10 = 0, appleOffset = 0;
  
  const cmap_encoding_subtable_header *esh;
  for (index = 0; index < nSubtables && msOffset10 == 0; ++index) {
    esh = &cmap->encoding_subtable_headers[index];
    switch (hb_be_uint16_t(esh->platform_id)) {
      case 3: // Microsoft encodings
	switch (hb_be_uint16_t(esh->encoding_id)) {
	  case 1:
	    msOffset1 = hb_be_uint32_t(esh->offset);
	    break;
	    
	  case 10:
	    msOffset10 = hb_be_uint32_t(esh->offset);
	    break;
	}
	break;
	
      case 0: // Apple Unicode table
	appleOffset = hb_be_uint32_t(esh->offset);
	break;		
    }
  }
  
  if (msOffset10 != 0) {
    subtable = (const guint16*)(cmap_data + msOffset10);
  } else if (msOffset1 != 0) {
    subtable = (const guint16*)(cmap_data + msOffset1);
  } else if (appleOffset != 0) {
    subtable = (const guint16*)(cmap_data + appleOffset);
  } else {
    return NULL;
  }
  
  guint16 format = hb_be_uint16_t(*subtable); // read the subtable format field
  switch (format) {
    case 4:
      if (subtable_length) {
	*subtable_length = hb_be_uint16_t(((const format_4_cmap*)subtable)->length);
      }
      return (const char*)subtable;
    case 12:
      if (subtable_length) {
	*subtable_length = hb_be_uint32_t(((const format_12_cmap*)subtable)->length);
      }
      return (const char*)subtable;
  }
  
  /* couldn't find a usable subtable */
  return NULL;
}

/*
 * Map the given character to glyph id using our preferred subtable of the given cmap.
 * This is a convenience function; but note that it is more efficient for the client to
 * get the preferred subtable and cache that, then call hb_cmap_subtable_char_to_glyph
 * to map each character.
 */
hb_codepoint_t
hb_cmap_char_to_glyph(const char *cmap_data, hb_codepoint_t unicode_scalar)
{
  const char *subtable = hb_cmap_find_preferred_subtable(cmap_data, NULL);
  if (subtable) {
    return hb_cmap_subtable_char_to_glyph(subtable, unicode_scalar);
  }
  else {
    return 0;
  }
}
