#include <ApplicationServices/ApplicationServices.h>

#include "hb-ot-layout.h"
#include "hb-script-itemize.h"
#include "hb-cmap.h"
#include "hb-script-code-map.h"
#include "hb-shape-generic.h"
#include "hb-shape-arabic.h"

static char*
getTable(ATSFontRef fontRef, UInt32 tag)
{
  ByteCount size;
  OSStatus status = ::ATSFontGetTable (fontRef, tag, 0, 0, NULL, &size);
  if (status == noErr) {
    char *table = new char[size];
    ::ATSFontGetTable(fontRef, tag, 0, size, table, &size);
    return table;
  }
  return NULL;
}

static CFStringRef
getFontForScript(hb_tag_t scriptTag)
{
  switch (scriptTag) {
    case 'arab': return CFSTR("TimesNewRomanPSMT");
    default: return CFSTR("GaramondPremrPro-It");
  }
}

void
process(const uint16_t *txt, guint32 length)
{
  hb_script_itemizer_t *scriptRun = hb_script_run_open(txt, length, HB_UTF16);
  uint32_t runStart, runLimit;
  hb_script_code_t runScript;

  while (hb_script_run_next(scriptRun, &runStart, &runLimit, &runScript)) {
    fprintf(stderr, "run [%d:%d], script %d\n", runStart, runLimit, runScript);

    const hb_script_code_mapping_t * script_info = hb_get_script_code_mapping(runScript);
    CFStringRef fontName = getFontForScript(script_info->script_tag);
    ATSFontRef fontRef = ::ATSFontFindFromPostScriptName(fontName,
							 kATSOptionFlagsDefault);

    char *gdef = getTable(fontRef, 'GDEF');
    char *gsub = getTable(fontRef, 'GSUB');
    char *gpos = getTable(fontRef, 'GPOS');
    char *cmap = getTable(fontRef, 'cmap');
    
    hb_ot_layout_t *layout = hb_ot_layout_create_for_tables (gdef, gsub, gpos);
    hb_ot_layout_set_scale(layout, 0x10000, 0x10000);
    hb_ot_layout_set_ppem(layout, 10, 10);
    hb_ot_layout_set_hinting(layout, 0);
    
    hb_buffer_t *buffer = hb_buffer_new(runLimit - runStart);
    while (runStart < runLimit) {
      /* only valid for BMP text at the moment! */
      hb_buffer_add_glyph(buffer, txt[runStart++], 0, 0);
    }
    
    // for experimenting...
    feature_setting_t f[] = {
      'swsh', 1,
    };
    int direction = 1;
    
    (script_info->shaping_function)(layout, buffer, cmap, script_info->shaping_data, direction, &f[0], sizeof(f) / sizeof(f[0]));
    
    for (int i = 0; i < buffer->in_length; ++i) {
      fprintf(stderr, "gindex = %d, pos = %d, adv = %d\n",
	    buffer->in_string[i].gindex,
	    buffer->positions ? buffer->positions[i].x_pos : 0,
	    buffer->positions ? buffer->positions[i].x_advance : 0);
    }
    
    hb_buffer_free(buffer);
    
    delete[] cmap;
    delete[] gdef;
    delete[] gsub;
    delete[] gpos;
  }
  hb_script_run_close(scriptRun);
}

int main (int argc, const char * argv[])
{
  uint16_t txt[] = {
    'T', 'e', 's', 't', ' ', '(', 0x014a, ')',
    0x0627, 0x0644, 0x0627, 0x0644, 0x0639, 0x0631, 0x0628, 0x064a,
    ' ', 'f', 'i', 'v', 'e', ' ', 'f', 'i', 'r', 'e', 'f', 'l', 'i', 'e', 's'
  };
  
  process(txt, sizeof(txt) / sizeof(txt[0]));

  return 0;
}
