/*
 * Copyright 2009 Mozilla Corporation.
 *
 * This is part of HarfBuzz, an OpenType Layout engine library.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Created by Jonathan Kew on 18/06/2009.
 */

#include "hb-script-code-map.h"
#include "hb-ot-layout.h"
#include "hb-shaper-private.h"
#include "hb-shape-generic.h"
#include "hb-shape-arabic.h"

/*
 * Array of OpenType script tags corresponding to ICU UScriptCode values.
 * Note that there must be no gaps in the sequence of assigned USCRIPT_...
 * values here, otherwise placeholder entries would need to be inserted
 * to keep the array indexing correct.
 */
static const
hb_script_code_mapping_t script_code_map[] = {
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_COMMON       =  0 , /* Zyyy */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_INHERITED    =  1,  /* Qaai */
  { hb_shape_arabic,  NULL, HB_TAG('a','r','a','b')         },    /* USCRIPT_ARABIC       =  2,  /* Arab */
  { hb_shape_generic, NULL, HB_TAG('a','r','m','n')         },    /* USCRIPT_ARMENIAN     =  3,  /* Armn */
  { hb_shape_generic, NULL, HB_TAG('b','n','g','2')         },    /* USCRIPT_BENGALI      =  4,  /* Beng */
  { hb_shape_generic, NULL, HB_TAG('b','o','p','o')         },    /* USCRIPT_BOPOMOFO     =  5,  /* Bopo */
  { hb_shape_generic, NULL, HB_TAG('c','h','e','r')         },    /* USCRIPT_CHEROKEE     =  6,  /* Cher */
  { hb_shape_generic, NULL, HB_TAG('c','o','p','t')         },    /* USCRIPT_COPTIC       =  7,  /* Copt */
  { hb_shape_generic, NULL, HB_TAG('c','y','r','l')         },    /* USCRIPT_CYRILLIC     =  8,  /* Cyrl */
  { hb_shape_generic, NULL, HB_TAG('d','s','r','t')         },    /* USCRIPT_DESERET      =  9,  /* Dsrt */
  { hb_shape_generic, NULL, HB_TAG('d','e','v','2')         },    /* USCRIPT_DEVANAGARI   = 10,  /* Deva */
  { hb_shape_generic, NULL, HB_TAG('e','t','h','i')         },    /* USCRIPT_ETHIOPIC     = 11,  /* Ethi */
  { hb_shape_generic, NULL, HB_TAG('g','e','o','r')         },    /* USCRIPT_GEORGIAN     = 12,  /* Geor */
  { hb_shape_generic, NULL, HB_TAG('g','o','t','h')         },    /* USCRIPT_GOTHIC       = 13,  /* Goth */
  { hb_shape_generic, NULL, HB_TAG('g','r','e','k')         },    /* USCRIPT_GREEK        = 14,  /* Grek */
  { hb_shape_generic, NULL, HB_TAG('g','j','r','2')         },    /* USCRIPT_GUJARATI     = 15,  /* Gujr */
  { hb_shape_generic, NULL, HB_TAG('g','u','r','2')         },    /* USCRIPT_GURMUKHI     = 16,  /* Guru */
  { hb_shape_generic, NULL, HB_TAG('h','a','n','i')         },    /* USCRIPT_HAN          = 17,  /* Hani */
  { hb_shape_generic, NULL, HB_TAG('h','a','n','g')         },    /* USCRIPT_HANGUL       = 18,  /* Hang */
  { hb_shape_generic, NULL, HB_TAG('h','e','b','r')         },    /* USCRIPT_HEBREW       = 19,  /* Hebr */
  { hb_shape_generic, NULL, HB_TAG('k','a','n','a')         },    /* USCRIPT_HIRAGANA     = 20,  /* Hira */
  { hb_shape_generic, NULL, HB_TAG('k','n','d','2')         },    /* USCRIPT_KANNADA      = 21,  /* Knda */
  { hb_shape_generic, NULL, HB_TAG('k','a','n','a')         },    /* USCRIPT_KATAKANA     = 22,  /* Kana */
  { hb_shape_generic, NULL, HB_TAG('k','h','m','r')         },    /* USCRIPT_KHMER        = 23,  /* Khmr */
  { hb_shape_generic, NULL, HB_TAG('l','a','o',' ')         },    /* USCRIPT_LAO          = 24,  /* Laoo */
  { hb_shape_generic, NULL, HB_TAG('l','a','t','n')         },    /* USCRIPT_LATIN        = 25,  /* Latn */
  { hb_shape_generic, NULL, HB_TAG('m','l','m','2')         },    /* USCRIPT_MALAYALAM    = 26,  /* Mlym */
  { hb_shape_generic, NULL, HB_TAG('m','o','n','g')         },    /* USCRIPT_MONGOLIAN    = 27,  /* Mong */
  { hb_shape_generic, NULL, HB_TAG('m','y','m','r')         },    /* USCRIPT_MYANMAR      = 28,  /* Mymr */
  { hb_shape_generic, NULL, HB_TAG('o','g','a','m')         },    /* USCRIPT_OGHAM        = 29,  /* Ogam */
  { hb_shape_generic, NULL, HB_TAG('i','t','a','l')         },    /* USCRIPT_OLD_ITALIC   = 30,  /* Ital */
  { hb_shape_generic, NULL, HB_TAG('o','r','y','2')         },    /* USCRIPT_ORIYA        = 31,  /* Orya */
  { hb_shape_generic, NULL, HB_TAG('r','u','n','r')         },    /* USCRIPT_RUNIC        = 32,  /* Runr */
  { hb_shape_generic, NULL, HB_TAG('s','i','n','h')         },    /* USCRIPT_SINHALA      = 33,  /* Sinh */
  { hb_shape_arabic,  NULL, HB_TAG('s','y','r','c')         },    /* USCRIPT_SYRIAC       = 34,  /* Syrc */
  { hb_shape_generic, NULL, HB_TAG('t','m','l','2')         },    /* USCRIPT_TAMIL        = 35,  /* Taml */
  { hb_shape_generic, NULL, HB_TAG('t','e','l','2')         },    /* USCRIPT_TELUGU       = 36,  /* Telu */
  { hb_shape_generic, NULL, HB_TAG('t','h','a','a')         },    /* USCRIPT_THAANA       = 37,  /* Thaa */
  { hb_shape_generic, NULL, HB_TAG('t','h','a','i')         },    /* USCRIPT_THAI         = 38,  /* Thai */
  { hb_shape_generic, NULL, HB_TAG('t','i','b','t')         },    /* USCRIPT_TIBETAN      = 39,  /* Tibt */
    /** Canadian_Aboriginal script. @stable ICU 2.6 */
  { hb_shape_generic, NULL, HB_TAG('c','a','n','s')         },    /* USCRIPT_CANADIAN_ABORIGINAL = 40,  /* Cans */
    /** Canadian_Aboriginal script (alias). @stable ICU 2.2 */
    /* USCRIPT_UCAS         = USCRIPT_CANADIAN_ABORIGINAL, */
  { hb_shape_generic, NULL, HB_TAG('y','i',' ',' ')         },    /* USCRIPT_YI           = 41,  /* Yiii */
  { hb_shape_generic, NULL, HB_TAG('t','g','l','g')         },    /* USCRIPT_TAGALOG      = 42,  /* Tglg */
  { hb_shape_generic, NULL, HB_TAG('h','a','n','o')         },    /* USCRIPT_HANUNOO      = 43,  /* Hano */
  { hb_shape_generic, NULL, HB_TAG('b','u','h','d')         },    /* USCRIPT_BUHID        = 44,  /* Buhd */
  { hb_shape_generic, NULL, HB_TAG('t','a','g','b')         },    /* USCRIPT_TAGBANWA     = 45,  /* Tagb */

    /* New scripts in Unicode 4 @stable ICU 2.6 */
  { hb_shape_generic, NULL, HB_TAG('b','r','a','i')         },    /* USCRIPT_BRAILLE      = 46,  /* Brai */
  { hb_shape_generic, NULL, HB_TAG('c','p','r','t')         },    /* USCRIPT_CYPRIOT      = 47,  /* Cprt */
  { hb_shape_generic, NULL, HB_TAG('l','i','m','b')         },    /* USCRIPT_LIMBU        = 48,  /* Limb */
  { hb_shape_generic, NULL, HB_TAG('l','i','n','b')         },    /* USCRIPT_LINEAR_B     = 49,  /* Linb */
  { hb_shape_generic, NULL, HB_TAG('o','s','m','a')         },    /* USCRIPT_OSMANYA      = 50,  /* Osma */
  { hb_shape_generic, NULL, HB_TAG('s','h','a','w')         },    /* USCRIPT_SHAVIAN      = 51,  /* Shaw */
  { hb_shape_generic, NULL, HB_TAG('t','a','l','e')         },    /* USCRIPT_TAI_LE       = 52,  /* Tale */
  { hb_shape_generic, NULL, HB_TAG('u','g','a','r')         },    /* USCRIPT_UGARITIC     = 53,  /* Ugar */

    /** New script code in Unicode 4.0.1 @stable ICU 3.0 */
  { hb_shape_generic, NULL, HB_TAG('k','a','n','a')         },    /* USCRIPT_KATAKANA_OR_HIRAGANA = 54,/*Hrkt */

    /* New scripts in Unicode 4.1 @stable ICU 3.4 */
  { hb_shape_generic, NULL, HB_TAG('b','u','g','i')         },    /* USCRIPT_BUGINESE      = 55, /* Bugi */
  { hb_shape_generic, NULL, HB_TAG('g','l','a','g')         },    /* USCRIPT_GLAGOLITIC    = 56, /* Glag */
  { hb_shape_generic, NULL, HB_TAG('k','h','a','r')         },    /* USCRIPT_KHAROSHTHI    = 57, /* Khar */
  { hb_shape_generic, NULL, HB_TAG('s','y','l','o')         },    /* USCRIPT_SYLOTI_NAGRI  = 58, /* Sylo */
  { hb_shape_generic, NULL, HB_TAG('t','a','l','u')         },    /* USCRIPT_NEW_TAI_LUE   = 59, /* Talu */
  { hb_shape_generic, NULL, HB_TAG('t','f','n','g')         },    /* USCRIPT_TIFINAGH      = 60, /* Tfng */
  { hb_shape_generic, NULL, HB_TAG('x','p','e','o')         },    /* USCRIPT_OLD_PERSIAN   = 61, /* Xpeo */

    /* New script codes from ISO 15924 @stable ICU 3.6 */
  { hb_shape_generic, NULL, HB_TAG('b','a','l','i')         },    /* USCRIPT_BALINESE                      = 62, /* Bali */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_BATAK                         = 63, /* Batk */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_BLISSYMBOLS                   = 64, /* Blis */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_BRAHMI                        = 65, /* Brah */
  { hb_shape_generic, NULL, HB_TAG('c','h','a','m')         },    /* USCRIPT_CHAM                          = 66, /* Cham */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_CIRTH                         = 67, /* Cirt */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_OLD_CHURCH_SLAVONIC_CYRILLIC  = 68, /* Cyrs */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_DEMOTIC_EGYPTIAN              = 69, /* Egyd */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_HIERATIC_EGYPTIAN             = 70, /* Egyh */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_EGYPTIAN_HIEROGLYPHS          = 71, /* Egyp */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_KHUTSURI                      = 72, /* Geok */
  { hb_shape_generic, NULL, HB_TAG('h','a','n','i')         },    /* USCRIPT_SIMPLIFIED_HAN                = 73, /* Hans */
  { hb_shape_generic, NULL, HB_TAG('h','a','n','i')         },    /* USCRIPT_TRADITIONAL_HAN               = 74, /* Hant */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_PAHAWH_HMONG                  = 75, /* Hmng */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_OLD_HUNGARIAN                 = 76, /* Hung */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_HARAPPAN_INDUS                = 77, /* Inds */
  { hb_shape_generic, NULL, HB_TAG('j','a','v','a')         },    /* USCRIPT_JAVANESE                      = 78, /* Java */
  { hb_shape_generic, NULL, HB_TAG('k','a','l','i')         },    /* USCRIPT_KAYAH_LI                      = 79, /* Kali */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_LATIN_FRAKTUR                 = 80, /* Latf */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_LATIN_GAELIC                  = 81, /* Latg */
  { hb_shape_generic, NULL, HB_TAG('l','e','p','c')         },    /* USCRIPT_LEPCHA                        = 82, /* Lepc */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_LINEAR_A                      = 83, /* Lina */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_MANDAEAN                      = 84, /* Mand */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_MAYAN_HIEROGLYPHS             = 85, /* Maya */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_MEROITIC                      = 86, /* Mero */
  { hb_shape_arabic,  NULL, HB_TAG('n','k','o',' ')         },    /* USCRIPT_NKO                           = 87, /* Nkoo */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_ORKHON                        = 88, /* Orkh */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_OLD_PERMIC                    = 89, /* Perm */
  { hb_shape_generic, NULL, HB_TAG('p','h','a','g')         },    /* USCRIPT_PHAGS_PA                      = 90, /* Phag */
  { hb_shape_generic, NULL, HB_TAG('p','h','n','x')         },    /* USCRIPT_PHOENICIAN                    = 91, /* Phnx */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_PHONETIC_POLLARD              = 92, /* Plrd */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_RONGORONGO                    = 93, /* Roro */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_SARATI                        = 94, /* Sara */
  { hb_shape_generic, NULL, HB_TAG('s','y','r','c')         },    /* USCRIPT_ESTRANGELO_SYRIAC             = 95, /* Syre */
  { hb_shape_generic, NULL, HB_TAG('s','y','r','c')         },    /* USCRIPT_WESTERN_SYRIAC                = 96, /* Syrj */
  { hb_shape_generic, NULL, HB_TAG('s','y','r','c')         },    /* USCRIPT_EASTERN_SYRIAC                = 97, /* Syrn */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_TENGWAR                       = 98, /* Teng */
  { hb_shape_generic, NULL, HB_TAG('v','a','i',' ')         },    /* USCRIPT_VAI                           = 99, /* Vaii */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_VISIBLE_SPEECH                = 100, /* Visp */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_CUNEIFORM                     = 101,/* Xsux */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_UNWRITTEN_LANGUAGES           = 102,/* Zxxx */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_UNKNOWN                       = 103,/* Zzzz */ /* Unknown="Code for uncoded script", for unassigned code points */

    /* New script codes from ISO 15924 @stable ICU 3.8 */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_CARIAN                        = 104,/* Cari */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_JAPANESE                      = 105,/* Jpan */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_LANNA                         = 106,/* Lana */
  { hb_shape_generic, NULL, HB_TAG('l','y','c','i')         },    /* USCRIPT_LYCIAN                        = 107,/* Lyci */
  { hb_shape_generic, NULL, HB_TAG('l','y','d','i')         },    /* USCRIPT_LYDIAN                        = 108,/* Lydi */
  { hb_shape_generic, NULL, HB_TAG('o','l','c','k')         },    /* USCRIPT_OL_CHIKI                      = 109,/* Olck */
  { hb_shape_generic, NULL, HB_TAG('r','j','n','g')         },    /* USCRIPT_REJANG                        = 110,/* Rjng */
  { hb_shape_generic, NULL, HB_TAG('s','a','u','r')         },    /* USCRIPT_SAURASHTRA                    = 111,/* Saur */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_SIGN_WRITING                  = 112,/* Sgnw */
  { hb_shape_generic, NULL, HB_TAG('s','u','n','d')         },    /* USCRIPT_SUNDANESE                     = 113,/* Sund */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_MOON                          = 114,/* Moon */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_MEITEI_MAYEK                  = 115,/* Mtei */

    /* New script codes from ISO 15924 @stable ICU 4.0 */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_IMPERIAL_ARAMAIC              = 116,/* Armi */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_AVESTAN                       = 117,/* Avst */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_CHAKMA                        = 118,/* Cakm */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_KOREAN                        = 119,/* Kore */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_KAITHI                        = 120,/* Kthi */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_MANICHAEAN                    = 121,/* Mani */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_INSCRIPTIONAL_PAHLAVI         = 122,/* Phli */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_PSALTER_PAHLAVI               = 123,/* Phlp */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_BOOK_PAHLAVI                  = 124,/* Phlv */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_INSCRIPTIONAL_PARTHIAN        = 125,/* Prti */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_SAMARITAN                     = 126,/* Samr */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT },    /* USCRIPT_TAI_VIET                      = 127,/* Tavt */
  { hb_shape_generic, NULL, HB_TAG('m','a','t','h')         },    /* USCRIPT_MATHEMATICAL_NOTATION         = 128,/* Zmth */
  { hb_shape_generic, NULL, HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT }     /* USCRIPT_SYMBOLS                       = 129,/* Zsym */

    /* Private use codes from Qaaa - Qabx are not supported*/
    /* USCRIPT_CODE_LIMIT    = 130 */
};

hb_tag_t
hb_get_old_script_tag(hb_tag_t new_tag)
{
  switch (new_tag) {
    case HB_TAG('b','n','g','2'): return HB_TAG('b','e','n','g');
    case HB_TAG('d','e','v','2'): return HB_TAG('d','e','v','a');
    case HB_TAG('g','j','r','2'): return HB_TAG('g','u','j','r');
    case HB_TAG('g','u','r','2'): return HB_TAG('g','u','r','u');
    case HB_TAG('k','n','d','2'): return HB_TAG('k','n','d','a');
    case HB_TAG('m','l','m','2'): return HB_TAG('m','l','y','m');
    case HB_TAG('o','r','y','2'): return HB_TAG('o','r','y','a');
    case HB_TAG('t','m','l','2'): return HB_TAG('t','a','m','l');
    case HB_TAG('t','e','l','2'): return HB_TAG('t','e','l','u');
    default:                      return HB_OT_LAYOUT_TAG_DEFAULT_SCRIPT;
  }
}

const hb_script_code_mapping_t *
hb_get_script_code_mapping(hb_script_code_t script_code)
{
  uint32_t index = script_code;
  if (index < sizeof(script_code_map) / sizeof(script_code_map[0])) {
    return &script_code_map[index];
  }
  /* in case of error, return first entry, which is a "generic" one */
  return &script_code_map[0];
}
