/*
 * Copyright 2009 Mozilla Corporation.
 *
 * This is part of HarfBuzz, an OpenType Layout engine library.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Created by Jonathan Kew on 23/06/2009.
 *
 * This file is based on usc_impl.c from ICU 4.2.0.1, slightly adapted
 * for use as part of Harfbuzz, separate from a standard ICU build.
 *
 * The original ICU license of the code follows:
 *
 * ICU License - ICU 1.8.1 and later
 *
 * COPYRIGHT AND PERMISSION NOTICE
 * 
 * Copyright (c) 1995-2009 International Business Machines Corporation and
 * others
 *
 * All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, provided that the above copyright notice(s) and this
 * permission notice appear in all copies of the Software and that both the
 * above copyright notice(s) and this permission notice appear in supporting
 * documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE
 * BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Except as contained in this notice, the name of a copyright holder shall
 * not be used in advertising or otherwise to promote the sale, use or other
 * dealings in this Software without prior written authorization of the
 * copyright holder.
 *
 * All trademarks and registered trademarks mentioned herein are the property
 * of their respective owners. 
 */

#include "hb-script-itemize.h"
#include "hb-unicode-properties.h"

#include <stdlib.h>

#define ARRAY_SIZE(array) (sizeof array  / sizeof array[0])

#define PAREN_STACK_DEPTH 32

#define MOD(sp) ((sp) % PAREN_STACK_DEPTH)
#define LIMIT_INC(sp) (((sp) < PAREN_STACK_DEPTH)? (sp) + 1 : PAREN_STACK_DEPTH)
#define INC(sp,count) (MOD((sp) + (count)))
#define INC1(sp) (INC(sp, 1))
#define DEC(sp,count) (MOD((sp) + PAREN_STACK_DEPTH - (count)))
#define DEC1(sp) (DEC(sp, 1))
#define STACK_IS_EMPTY(scriptRun) ((scriptRun)->pushCount <= 0)
#define STACK_IS_NOT_EMPTY(scriptRun) (! STACK_IS_EMPTY(scriptRun))
#define TOP(scriptRun) ((scriptRun)->parenStack[(scriptRun)->parenSP])
#define SYNC_FIXUP(scriptRun) ((scriptRun)->fixupCount = 0)

struct ParenStackEntry
{
  int32_t pairIndex;
  hb_script_code_t scriptCode;
};

struct _hb_script_itemizer_t
{
  uint32_t textLength;
  const void *textPtr;
  
  uint32_t scriptStart;
  uint32_t scriptLimit;
  hb_script_code_t scriptCode;
  
  hb_text_encoding_form_t encodingForm;
  
  struct ParenStackEntry parenStack[PAREN_STACK_DEPTH];
  uint32_t parenSP;
  uint32_t pushCount;
  uint32_t fixupCount;
};

static uint8_t highBit(uint32_t value);

static const uint16_t pairedChars[] = {
  0x0028, 0x0029, /* ascii paired punctuation */
  0x003c, 0x003e,
  0x005b, 0x005d,
  0x007b, 0x007d,
  0x00ab, 0x00bb, /* guillemets */
  0x2018, 0x2019, /* general punctuation */
  0x201c, 0x201d,
  0x2039, 0x203a,
  0x3008, 0x3009, /* chinese paired punctuation */
  0x300a, 0x300b,
  0x300c, 0x300d,
  0x300e, 0x300f,
  0x3010, 0x3011,
  0x3014, 0x3015,
  0x3016, 0x3017,
  0x3018, 0x3019,
  0x301a, 0x301b
};

static void push(hb_script_itemizer_t * scriptRun, int32_t pairIndex, hb_script_code_t scriptCode)
{
  scriptRun->pushCount  = LIMIT_INC(scriptRun->pushCount);
  scriptRun->fixupCount = LIMIT_INC(scriptRun->fixupCount);
  
  scriptRun->parenSP = INC1(scriptRun->parenSP);
  scriptRun->parenStack[scriptRun->parenSP].pairIndex  = pairIndex;
  scriptRun->parenStack[scriptRun->parenSP].scriptCode = scriptCode;
}

static void pop(hb_script_itemizer_t *scriptRun)
{
  if (STACK_IS_EMPTY(scriptRun)) {
    return;
  }
  
  if (scriptRun->fixupCount > 0) {
    scriptRun->fixupCount -= 1;
  }
  
  scriptRun->pushCount -= 1;
  scriptRun->parenSP = DEC1(scriptRun->parenSP);
  
  /* If the stack is now empty, reset the stack
   pointers to their initial values.
   */
  if (STACK_IS_EMPTY(scriptRun)) {
    scriptRun->parenSP = -1;
  }
}

static void fixup(hb_script_itemizer_t *scriptRun, hb_script_code_t scriptCode)
{
  int32_t fixupSP = DEC(scriptRun->parenSP, scriptRun->fixupCount);
  
  while (scriptRun->fixupCount-- > 0) {
    fixupSP = INC1(fixupSP);
    scriptRun->parenStack[fixupSP].scriptCode = scriptCode;
  }
}

static uint32_t
find_highest_bit(uint32_t value)
{
  /* propagate highest bit into all lower bits of the value */
  value |= (value >> 1);
  value |= (value >> 2);
  value |= (value >> 4);
  value |= (value >> 8);
  value |= (value >> 16);
  /* isolate the leftmost bit */
  return (value & ~(value >> 1));
}

static int32_t
getPairIndex(hb_codepoint_t ch)
{
  int32_t pairedCharCount = ARRAY_SIZE(pairedChars);
  int32_t pairedCharPower = find_highest_bit(pairedCharCount);
  int32_t pairedCharExtra = pairedCharCount - pairedCharPower;
  
  int32_t probe = pairedCharPower;
  int32_t pairIndex = 0;
  
  if (ch >= pairedChars[pairedCharExtra]) {
    pairIndex = pairedCharExtra;
  }
  
  while (probe > 1) {
    probe >>= 1;
    
    if (ch >= pairedChars[pairIndex + probe]) {
      pairIndex += probe;
    }
  }
  
  if (pairedChars[pairIndex] != ch) {
    pairIndex = -1;
  }
  
  return pairIndex;
}

static int
sameScript(hb_script_code_t scriptOne, hb_script_code_t scriptTwo)
{
  return scriptOne <= HB_SCRIPT_INHERITED || scriptTwo <= HB_SCRIPT_INHERITED || scriptOne == scriptTwo;
}

hb_script_itemizer_t *
hb_script_run_open(const void *src, uint32_t length, hb_text_encoding_form_t encoding)
{
  hb_script_itemizer_t * result = (hb_script_itemizer_t *) malloc(sizeof(hb_script_itemizer_t));
  
  if (result == NULL) {
    return NULL;
  }
  
  hb_script_run_set_text(result, src, length, encoding);
  
  return result;
}

void
hb_script_run_close(hb_script_itemizer_t *scriptRun)
{
  if (scriptRun != NULL) {
    free(scriptRun);
  }
}

void
hb_script_run_reset(hb_script_itemizer_t * scriptRun)
{
  if (scriptRun != NULL) {
    scriptRun->scriptStart = 0;
    scriptRun->scriptLimit = 0;
    scriptRun->scriptCode  = HB_SCRIPT_INVALID_CODE;
    scriptRun->parenSP     = -1;
    scriptRun->pushCount   =  0;
    scriptRun->fixupCount  =  0;
  }
}

void
hb_script_run_set_text(hb_script_itemizer_t * scriptRun, const void * src, uint32_t length, hb_text_encoding_form_t encoding)
{
  if (scriptRun == NULL) {
    return;
  }
  
  scriptRun->textPtr  = src;
  scriptRun->textLength = length;
  scriptRun->encodingForm = encoding;
  
  hb_script_run_reset(scriptRun);
}

/*
 * The following is based on ICU, primarily the U8_NEXT_UNSAFE macro
 * from utf8.h and the utf8_countTrailBytes table it uses.
 * Does not check for invalid UTF-8 sequences; will return garbled
 * Unicode character values for invalid input.
 */
 
static const uint8_t 
utf8_countTrailBytes[256]={
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,

    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    3, 3, 3, 3, 3,
    3, 3, 3,    /* illegal in Unicode */
    4, 4, 4, 4, /* illegal in Unicode */
    5, 5,       /* illegal in Unicode */
    0, 0        /* illegal bytes 0xfe and 0xff */
};

/**
 * Count the trail bytes for a UTF-8 lead byte.
 * @internal
 */
#define U8_COUNT_TRAIL_BYTES(leadByte) (utf8_countTrailBytes[(uint8_t)leadByte])

/**
 * Mask a UTF-8 lead byte, leave only the lower bits that form part of the code point value.
 * @internal
 */
#define U8_MASK_LEAD_BYTE(leadByte, countTrailBytes) ((leadByte)&=(1<<(6-(countTrailBytes)))-1)

/**
 * Get a code point from a string at a code point boundary offset,
 * and advance the offset to the next code point boundary.
 * (Post-incrementing forward iteration.)
 * "Unsafe" macro, assumes well-formed UTF-8.
 *
 * The offset may point to the lead byte of a multi-byte sequence,
 * in which case the macro will read the whole sequence.
 * The result is undefined if the offset points to a trail byte
 * or an illegal UTF-8 sequence.
 *
 * @param s const uint8_t * string
 * @param i string offset
 * @param l string length in code units
 * @param c output uint32_t variable
 */
#define UTF8_GET_NEXT_CHAR(s, i, l, c) { \
    (c)=(uint8_t)(s)[(i)++]; \
    if((uint8_t)((c)-0xc0)<0x35) { \
        uint8_t __count=U8_COUNT_TRAIL_BYTES(c); \
	if ((i) + __count > (l)) { \
	  /* run out of text: safely end reading, returning REPLACEMENT CHARACTER */ \
	  (i) = (l); \
	  (c) = 0xfffd; \
	} else { \
	  U8_MASK_LEAD_BYTE(c, __count); \
	  switch(__count) { \
	  /* each following branch falls through to the next one */ \
	  case 3: \
	      (c)=((c)<<6)|((s)[(i)++]&0x3f); \
	  case 2: \
	      (c)=((c)<<6)|((s)[(i)++]&0x3f); \
	  case 1: \
	      (c)=((c)<<6)|((s)[(i)++]&0x3f); \
	  /* no other branches to optimize switch() */ \
	      break; \
	  } \
       } \
    } \
}

int
hb_script_run_next(hb_script_itemizer_t *scriptRun, uint32_t *pRunStart, uint32_t *pRunLimit, hb_script_code_t *pRunScript)
{
  /* if we've fallen off the end of the text, we're done */
  if (scriptRun == NULL || scriptRun->scriptLimit >= scriptRun->textLength) {
    return 0;
  }
  
  SYNC_FIXUP(scriptRun);
  scriptRun->scriptCode = HB_SCRIPT_COMMON;
  
  for (scriptRun->scriptStart = scriptRun->scriptLimit;
       scriptRun->scriptLimit < scriptRun->textLength;
       scriptRun->scriptLimit += 1) {
    hb_codepoint_t ch;
    hb_script_code_t sc;
    int32_t pairIndex;
    uint32_t startOfChar = scriptRun->scriptLimit;

    switch (scriptRun->encodingForm) {
      case HB_UTF8: {
	/* get character code from UTF-8 buffer */
	const uint8_t * textArray = (const uint8_t *) scriptRun->textPtr;
	UTF8_GET_NEXT_CHAR(textArray, scriptRun->scriptLimit, scriptRun->textLength, ch);
	break;
      }
	
      case HB_UTF16: {
	/* decode UTF-16 (may be surrogate pair) */
	const uint16_t * textArray = (const uint16_t *) scriptRun->textPtr;
	ch = textArray[scriptRun->scriptLimit];
	if (ch >= 0xD800 && ch <= 0xDBFF && scriptRun->scriptLimit < scriptRun->textLength - 1) {
	  hb_codepoint_t low = textArray[scriptRun->scriptLimit + 1];
	  if (low >= 0xDC00 && low <= 0xDFFF) {
	    ch = (ch - 0xD800) * 0x0400 + low - 0xDC00 + 0x10000;
	    scriptRun->scriptLimit += 1;
	  }
	}
	break;
      }
	
      case HB_UTF32: {
	const uint32_t * textArray = (const uint32_t *)scriptRun->textPtr;
	ch = textArray[scriptRun->scriptLimit];
	break;
      }
    }
    
    sc = hb_get_script_code(ch);
    pairIndex = getPairIndex(ch);
    
    /*
     * Paired character handling:
     *
     * if it's an open character, push it onto the stack.
     * if it's a close character, find the matching open on the
     * stack, and use that script code. Any non-matching open
     * characters above it on the stack will be poped.
     */
    if (pairIndex >= 0) {
      if ((pairIndex & 1) == 0) {
	push(scriptRun, pairIndex, scriptRun->scriptCode);
      } else {
	int32_t pi = pairIndex & ~1;
	
	while (STACK_IS_NOT_EMPTY(scriptRun) && TOP(scriptRun).pairIndex != pi) {
	  pop(scriptRun);
	}
	
	if (STACK_IS_NOT_EMPTY(scriptRun)) {
	  sc = TOP(scriptRun).scriptCode;
	}
      }
    }
    
    if (sameScript(scriptRun->scriptCode, sc)) {
      if (scriptRun->scriptCode <= HB_SCRIPT_INHERITED && sc > HB_SCRIPT_INHERITED) {
	scriptRun->scriptCode = sc;
	
	fixup(scriptRun, scriptRun->scriptCode);
      }
      
      /*
       * if this character is a close paired character,
       * pop the matching open character from the stack
       */
      if (pairIndex >= 0 && (pairIndex & 1) != 0) {
	pop(scriptRun);
      }
    } else {
      /*
       * reset scriptLimit in case it was advanced during reading a
       * multiple-code-unit character
       */
      scriptRun->scriptLimit = startOfChar;

      break;
    }
  }
  
  if (pRunStart != NULL) {
    *pRunStart = scriptRun->scriptStart;
  }
  if (pRunLimit != NULL) {
    *pRunLimit = scriptRun->scriptLimit;
  }
  if (pRunScript != NULL) {
    *pRunScript = scriptRun->scriptCode;
  }
  
  return 1;
}
