#!/usr/bin/env perl

# argument is the directory where current UCD files are saved
$ucd = $ARGV[0];

# split 16-bit values into two fields, for "page" and "char" indexing
@bmpsplit = (10, 6);

$indexLen = 1 << $bmpsplit[0];
$dataLen = 1 << $bmpsplit[1];

# values from enum UJoiningType in uchar.h,
# plus values for special Syriac chars
%joiningType = (
  "U", 0,
  "C", 1,
  "D", 2,
  "L", 3,
  "R", 4,
  "T", 5,
  "Alaph", 6,
  "DalathRish", 7
);

# values from enum UCharDirection in uchar.h
%bidiClass = (
  "L", 0,
  "R", 1,
  "EN", 2,
  "ES", 3,
  "ET", 4,
  "AN", 5,
  "CS", 6,
  "B", 7,
  "S", 8,
  "WS", 9,
  "ON", 10,
  "LRE", 11,
  "LRO", 12,
  "AL", 13,
  "RLE", 14,
  "RLO", 15,
  "PDF", 16,
  "NSM", 17,
  "BN", 18
);

# load mapping of script names to UScript codes
while (<DATA>) {
    if (m/USCRIPT_([A-Z_]+)\s*=\s*(\d+)\s*,/) {
        $scriptCode{$1} = $2;
    }
}

# initialize default properties
for ($i = 0; $i < 0x110000; ++$i) {
    $join[$i] = $joiningType{"U"};
    $bidi[$i] = $bidiClass{"L"};
    $script[$i] = $scriptCode{"UNKNOWN"};
}
for ($i = 0x0600; $i <= 0x06ff; ++$i) { $bidi[$i] = $bidiClass{"AL"}; }
for ($i = 0x0700; $i <= 0x074f; ++$i) { $bidi[$i] = $bidiClass{"AL"}; }
for ($i = 0x0750; $i <= 0x077f; ++$i) { $bidi[$i] = $bidiClass{"AL"}; }
for ($i = 0x0780; $i <= 0x07bf; ++$i) { $bidi[$i] = $bidiClass{"AL"}; }
for ($i = 0xfb50; $i <= 0xfdff; ++$i) { $bidi[$i] = $bidiClass{"AL"}; }
for ($i = 0xfe70; $i <= 0xfeff; ++$i) { $bidi[$i] = $bidiClass{"AL"}; }


# read DerivedJoiningType.txt
open FH, "< $ucd/extracted/DerivedJoiningType.txt" or die "can't open DerivedJoiningType.txt";
while (<FH>) {
    if (m/([0-9A-F]{4,6})(?:\.\.([0-9A-F]{4,6}))*\s+;\s+(.)\s+#/) {
        $join = $joiningType{$3};
        $start = hex "0x$1";
        $end = (defined $2) ? hex "0x$2" : $start;
        for ($i = $start; $i <= $end; ++$i) {
            $join[$i] = $join;
        }
    }
}
close FH;


# we have special joining types for Syriac ALAPH and DALATH/RISH
$join[0x0710] = $joiningType{"Alaph"};
$join[0x0715] = $joiningType{"DalathRish"};
$join[0x0716] = $joiningType{"DalathRish"};
$join[0x072A] = $joiningType{"DalathRish"};
$join[0x072F] = $joiningType{"DalathRish"};


# read DerivedBidiClass.txt
open FH, "< $ucd/extracted/DerivedBidiClass.txt" or die "can't open DerivedBidiClass.txt";
while (<FH>) {
    if (m/([0-9A-F]{4,6})(?:\.\.([0-9A-F]{4,6}))*\s+;\s+(.)\s+#/) {
        $bidi = $bidiClass{$3};
        $start = hex "0x$1";
        $end = (defined $2) ? hex "0x$2" : $start;
        for ($i = $start; $i <= $end; ++$i) {
            $bidi[$i] = $bidi;
        }
    }
}
close FH;


# load mapping of script names to UScript codes
while (<DATA>) {
    if (m/USCRIPT_([A-Z_]+)\s*=\s*(\d+)\s*,/) {
        $scriptCode{$1} = $2;
    }
}

# read Scripts.txt
open FH, "< $ucd/Scripts.txt" or die "can't open Scripts.txt";
while (<FH>) {
    if (m/([0-9A-F]{4,6})(?:\.\.([0-9A-F]{4,6}))*\s+;\s+([^ ]+)/) {
        $script = uc($3);
        warn "unknown script $script" unless exists $scriptCode{$script};
        $script = $scriptCode{$script};
        $start = hex "0x$1";
        $end = (defined $2) ? hex "0x$2" : $start;
        for ($i = $start; $i <= $end; ++$i) {
            $script[$i] = $script;
        }
    }
}
close FH;

for ($i = 0; $i < 0x110000; ++$i) {
    $value[$i] = ($script[$i] << 8) + ($join[$i] << 5) + $bidi[$i];
}

$planeMap = "\x00" x 16;
foreach $plane (0 .. 16) {
	$pageMap = "\x00" x $indexLen * 2;
	foreach $page (0 .. $indexLen - 1) {
		$char = "\x00" x $dataLen * 2;
		foreach $ch (0 .. $dataLen - 1) {
            $usv = $plane * 0x10000 + $page * $dataLen + $ch;
			substr($char, $ch * 2, 2) = pack('S', $value[$usv]);
		}

		unless (exists $charindex{$char}) {
			$charindex{$char} = scalar keys %charindex;
			$char[$charindex{$char}] = $char;
		}
		substr($pageMap, $page * 2, 2) = pack('S', $charindex{$char});
	}
	
	unless (exists $pageMapIndex{$pageMap}) {
		$pageMapIndex{$pageMap} = scalar keys %pageMapIndex;
		$pageMap[$pageMapIndex{$pageMap}] = $pageMap;
	}
	if ($plane > 0) {
    	substr($planeMap, $plane - 1, 1) = pack('C', $pageMapIndex{$pageMap});
    }
}

$timestamp = gmtime();
print <<__END;
/*
 * * * * * This file contains MACHINE-GENERATED DATA, do not edit! * * * * *
 *
 * Copyright 2009 Mozilla Corporation.
 *
 * This is part of HarfBuzz, an OpenType Layout engine library.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Derived from the Unicode Character Database by hb-unicode-data-gen.pl
 *
 * For Unicode terms of use, see http://www.unicode.org/terms_of_use.html
 *
 * Created on $timestamp.
 *
 * * * * * This file contains MACHINE-GENERATED DATA, do not edit! * * * * *
 */

__END

print "#define INDEX_BITS $bmpsplit[0]\n";
print "#define CHAR_BITS $bmpsplit[1]\n";
print "\n";
print "static const uint8_t planes[16] = {";
print join(',', map { sprintf("%d", $_) } unpack('C*', $planeMap));
print "};\n\n";

$chCount = scalar @char;
$pmBits = $chCount > 255 ? 16 : 8;
$pmCount = scalar @pageMap;
print "static const uint${pmBits}_t pages[$pmCount][$indexLen] = {\n";
for ($i = 0; $i < scalar @pageMap; ++$i) {
	print "\t{";
	print join(',', map { sprintf("%d", $_) } unpack('S*', $pageMap[$i]));
	print $i < $#pageMap ? "},\n" : "}\n";
}
print "};\n\n";

print "typedef struct char_props { uint8_t script; uint8_t type_class; } char_props;\n";
print "static const char_props values[$chCount][$dataLen] = {\n";
for ($i = 0; $i < scalar @char; ++$i) {
	print "\t{";
	print join(',', map { sprintf("{%d,%d}", $_ >> 8, $_ & 0xff) } unpack('S*', $char[$i]));
	print $i < $#char ? "},\n" : "}\n";
}
print "};\n\n";

__DATA__
from uscript.h in ICU 4.2.0.1

      USCRIPT_COMMON       =  0 , /* Zyyy */
      USCRIPT_INHERITED    =  1,  /* Qaai */
      USCRIPT_ARABIC       =  2,  /* Arab */
      USCRIPT_ARMENIAN     =  3,  /* Armn */
      USCRIPT_BENGALI      =  4,  /* Beng */
      USCRIPT_BOPOMOFO     =  5,  /* Bopo */
      USCRIPT_CHEROKEE     =  6,  /* Cher */
      USCRIPT_COPTIC       =  7,  /* Copt */
      USCRIPT_CYRILLIC     =  8,  /* Cyrl */
      USCRIPT_DESERET      =  9,  /* Dsrt */
      USCRIPT_DEVANAGARI   = 10,  /* Deva */
      USCRIPT_ETHIOPIC     = 11,  /* Ethi */
      USCRIPT_GEORGIAN     = 12,  /* Geor */
      USCRIPT_GOTHIC       = 13,  /* Goth */
      USCRIPT_GREEK        = 14,  /* Grek */
      USCRIPT_GUJARATI     = 15,  /* Gujr */
      USCRIPT_GURMUKHI     = 16,  /* Guru */
      USCRIPT_HAN          = 17,  /* Hani */
      USCRIPT_HANGUL       = 18,  /* Hang */
      USCRIPT_HEBREW       = 19,  /* Hebr */
      USCRIPT_HIRAGANA     = 20,  /* Hira */
      USCRIPT_KANNADA      = 21,  /* Knda */
      USCRIPT_KATAKANA     = 22,  /* Kana */
      USCRIPT_KHMER        = 23,  /* Khmr */
      USCRIPT_LAO          = 24,  /* Laoo */
      USCRIPT_LATIN        = 25,  /* Latn */
      USCRIPT_MALAYALAM    = 26,  /* Mlym */
      USCRIPT_MONGOLIAN    = 27,  /* Mong */
      USCRIPT_MYANMAR      = 28,  /* Mymr */
      USCRIPT_OGHAM        = 29,  /* Ogam */
      USCRIPT_OLD_ITALIC   = 30,  /* Ital */
      USCRIPT_ORIYA        = 31,  /* Orya */
      USCRIPT_RUNIC        = 32,  /* Runr */
      USCRIPT_SINHALA      = 33,  /* Sinh */
      USCRIPT_SYRIAC       = 34,  /* Syrc */
      USCRIPT_TAMIL        = 35,  /* Taml */
      USCRIPT_TELUGU       = 36,  /* Telu */
      USCRIPT_THAANA       = 37,  /* Thaa */
      USCRIPT_THAI         = 38,  /* Thai */
      USCRIPT_TIBETAN      = 39,  /* Tibt */
      /** Canadian_Aboriginal script. @stable ICU 2.6 */
      USCRIPT_CANADIAN_ABORIGINAL = 40,  /* Cans */
      /** Canadian_Aboriginal script (alias). @stable ICU 2.2 */
      USCRIPT_UCAS         = USCRIPT_CANADIAN_ABORIGINAL,
      USCRIPT_YI           = 41,  /* Yiii */
      USCRIPT_TAGALOG      = 42,  /* Tglg */
      USCRIPT_HANUNOO      = 43,  /* Hano */
      USCRIPT_BUHID        = 44,  /* Buhd */
      USCRIPT_TAGBANWA     = 45,  /* Tagb */

      /* New scripts in Unicode 4 @stable ICU 2.6 */
      USCRIPT_BRAILLE      = 46,  /* Brai */
      USCRIPT_CYPRIOT      = 47,  /* Cprt */
      USCRIPT_LIMBU        = 48,  /* Limb */
      USCRIPT_LINEAR_B     = 49,  /* Linb */
      USCRIPT_OSMANYA      = 50,  /* Osma */
      USCRIPT_SHAVIAN      = 51,  /* Shaw */
      USCRIPT_TAI_LE       = 52,  /* Tale */
      USCRIPT_UGARITIC     = 53,  /* Ugar */

      /** New script code in Unicode 4.0.1 @stable ICU 3.0 */
      USCRIPT_KATAKANA_OR_HIRAGANA = 54,/*Hrkt */
      
      /* New scripts in Unicode 4.1 @stable ICU 3.4 */
      USCRIPT_BUGINESE      = 55, /* Bugi */
      USCRIPT_GLAGOLITIC    = 56, /* Glag */
      USCRIPT_KHAROSHTHI    = 57, /* Khar */
      USCRIPT_SYLOTI_NAGRI  = 58, /* Sylo */
      USCRIPT_NEW_TAI_LUE   = 59, /* Talu */
      USCRIPT_TIFINAGH      = 60, /* Tfng */
      USCRIPT_OLD_PERSIAN   = 61, /* Xpeo */

      /* New script codes from ISO 15924 @stable ICU 3.6 */
      USCRIPT_BALINESE                      = 62, /* Bali */
      USCRIPT_BATAK                         = 63, /* Batk */
      USCRIPT_BLISSYMBOLS                   = 64, /* Blis */
      USCRIPT_BRAHMI                        = 65, /* Brah */
      USCRIPT_CHAM                          = 66, /* Cham */
      USCRIPT_CIRTH                         = 67, /* Cirt */
      USCRIPT_OLD_CHURCH_SLAVONIC_CYRILLIC  = 68, /* Cyrs */
      USCRIPT_DEMOTIC_EGYPTIAN              = 69, /* Egyd */
      USCRIPT_HIERATIC_EGYPTIAN             = 70, /* Egyh */
      USCRIPT_EGYPTIAN_HIEROGLYPHS          = 71, /* Egyp */
      USCRIPT_KHUTSURI                      = 72, /* Geok */
      USCRIPT_SIMPLIFIED_HAN                = 73, /* Hans */
      USCRIPT_TRADITIONAL_HAN               = 74, /* Hant */
      USCRIPT_PAHAWH_HMONG                  = 75, /* Hmng */
      USCRIPT_OLD_HUNGARIAN                 = 76, /* Hung */
      USCRIPT_HARAPPAN_INDUS                = 77, /* Inds */
      USCRIPT_JAVANESE                      = 78, /* Java */
      USCRIPT_KAYAH_LI                      = 79, /* Kali */
      USCRIPT_LATIN_FRAKTUR                 = 80, /* Latf */
      USCRIPT_LATIN_GAELIC                  = 81, /* Latg */
      USCRIPT_LEPCHA                        = 82, /* Lepc */
      USCRIPT_LINEAR_A                      = 83, /* Lina */
      USCRIPT_MANDAEAN                      = 84, /* Mand */
      USCRIPT_MAYAN_HIEROGLYPHS             = 85, /* Maya */
      USCRIPT_MEROITIC                      = 86, /* Mero */
      USCRIPT_NKO                           = 87, /* Nkoo */
      USCRIPT_ORKHON                        = 88, /* Orkh */
      USCRIPT_OLD_PERMIC                    = 89, /* Perm */
      USCRIPT_PHAGS_PA                      = 90, /* Phag */
      USCRIPT_PHOENICIAN                    = 91, /* Phnx */
      USCRIPT_PHONETIC_POLLARD              = 92, /* Plrd */
      USCRIPT_RONGORONGO                    = 93, /* Roro */
      USCRIPT_SARATI                        = 94, /* Sara */
      USCRIPT_ESTRANGELO_SYRIAC             = 95, /* Syre */
      USCRIPT_WESTERN_SYRIAC                = 96, /* Syrj */
      USCRIPT_EASTERN_SYRIAC                = 97, /* Syrn */
      USCRIPT_TENGWAR                       = 98, /* Teng */
      USCRIPT_VAI                           = 99, /* Vaii */
      USCRIPT_VISIBLE_SPEECH                = 100, /* Visp */
      USCRIPT_CUNEIFORM                     = 101,/* Xsux */
      USCRIPT_UNWRITTEN_LANGUAGES           = 102,/* Zxxx */
      USCRIPT_UNKNOWN                       = 103,/* Zzzz */ /* Unknown="Code for uncoded script", for unassigned code points */

      /* New script codes from ISO 15924 @stable ICU 3.8 */
      USCRIPT_CARIAN                        = 104,/* Cari */
      USCRIPT_JAPANESE                      = 105,/* Jpan */
      USCRIPT_LANNA                         = 106,/* Lana */
      USCRIPT_LYCIAN                        = 107,/* Lyci */
      USCRIPT_LYDIAN                        = 108,/* Lydi */
      USCRIPT_OL_CHIKI                      = 109,/* Olck */
      USCRIPT_REJANG                        = 110,/* Rjng */
      USCRIPT_SAURASHTRA                    = 111,/* Saur */
      USCRIPT_SIGN_WRITING                  = 112,/* Sgnw */
      USCRIPT_SUNDANESE                     = 113,/* Sund */
      USCRIPT_MOON                          = 114,/* Moon */
      USCRIPT_MEITEI_MAYEK                  = 115,/* Mtei */

      /* New script codes from ISO 15924 @stable ICU 4.0 */
      USCRIPT_IMPERIAL_ARAMAIC              = 116,/* Armi */
      USCRIPT_AVESTAN                       = 117,/* Avst */
      USCRIPT_CHAKMA                        = 118,/* Cakm */
      USCRIPT_KOREAN                        = 119,/* Kore */
      USCRIPT_KAITHI                        = 120,/* Kthi */
      USCRIPT_MANICHAEAN                    = 121,/* Mani */
      USCRIPT_INSCRIPTIONAL_PAHLAVI         = 122,/* Phli */
      USCRIPT_PSALTER_PAHLAVI               = 123,/* Phlp */
      USCRIPT_BOOK_PAHLAVI                  = 124,/* Phlv */
      USCRIPT_INSCRIPTIONAL_PARTHIAN        = 125,/* Prti */
      USCRIPT_SAMARITAN                     = 126,/* Samr */
      USCRIPT_TAI_VIET                      = 127,/* Tavt */
      USCRIPT_MATHEMATICAL_NOTATION         = 128,/* Zmth */
      USCRIPT_SYMBOLS                       = 129,/* Zsym */
