/*
 * Copyright 2009 Mozilla Corporation.
 *
 * This is part of HarfBuzz, an OpenType Layout engine library.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Created by Jonathan Kew on 21/06/2009.
 */

#include "hb-shape-arabic.h"
#include "hb-shaper-private.h"
#include "hb-cmap.h"
#include "hb-ompl.h"
#include "hb-unicode-properties.h"

#define RTL_FEATURE_BIT  0x80000000
#define LTR_FEATURE_BIT  0x40000000

#define INIT_FEATURE_BIT 0x20000000
#define MEDI_FEATURE_BIT 0x10000000
#define FINA_FEATURE_BIT 0x08000000
#define ISOL_FEATURE_BIT 0x04000000
#define MED2_FEATURE_BIT 0x02000000
#define FIN2_FEATURE_BIT 0x01000000
#define FIN3_FEATURE_BIT 0x00800000

#define DEFAULT_MASK (INIT_FEATURE_BIT | MEDI_FEATURE_BIT | FINA_FEATURE_BIT | ISOL_FEATURE_BIT | MED2_FEATURE_BIT | FIN2_FEATURE_BIT | FIN3_FEATURE_BIT)

#define INITIAL_MASK (DEFAULT_MASK & ~INIT_FEATURE_BIT)
#define MEDIAL_MASK  (DEFAULT_MASK & ~MEDI_FEATURE_BIT)
#define FINAL_MASK   (DEFAULT_MASK & ~FINA_FEATURE_BIT)
#define ISOLATE_MASK (DEFAULT_MASK & ~ISOL_FEATURE_BIT)
#define MEDIAL2_MASK (DEFAULT_MASK & ~MED2_FEATURE_BIT)
#define FINAL2_MASK  (DEFAULT_MASK & ~FIN2_FEATURE_BIT)
#define FINAL3_MASK  (DEFAULT_MASK & ~FIN3_FEATURE_BIT)

struct st_entry {
  uint8_t next_state;
  uint8_t flags;
  uint8_t prev_mask_index;
  uint8_t curr_mask_index;
};
typedef struct st_entry st_entry;

static const uint32_t mask_array[] = {
  /* 0 */ 0, /* UNUSED: a zero mask_index in the state table means no change */
  /* 1 */ INITIAL_MASK,
  /* 2 */ MEDIAL_MASK,
  /* 3 */ FINAL_MASK,
  /* 4 */ ISOLATE_MASK,
  /* 5 */ MEDIAL2_MASK,
  /* 6 */ FINAL2_MASK,
  /* 7 */ FINAL3_MASK,
  /* 8 */ DEFAULT_MASK
};

#define NUM_STATES 8

static const st_entry
arabic_state_table[HB_NUM_JOIN_TYPES * NUM_STATES] =
{
  /*jt_NONE, jt_CAUS,   jt_DUAL,   jt_LEFT,   jt_RGHT,   jt_TRAN,   jt_ALPH,   jt_DLRS*/
  /* 0: start */
  {0,0,0,8}, {1,0,0,8}, {3,1,0,4}, {2,1,0,4}, {5,0,0,4}, {0,0,0,8}, {5,0,0,4}, {6,0,0,4},

  /* 1: seen join-causing */
  {0,0,0,8}, {1,0,0,8}, {4,1,0,3}, {2,1,0,4}, {5,0,0,3}, {1,0,0,8}, {7,1,0,3}, {6,0,0,3},

  /* 2: marked left-joining */
  {0,0,0,8}, {1,0,1,8}, {4,1,1,3}, {2,1,0,4}, {5,0,1,3}, {2,0,0,8}, {7,1,1,3}, {6,0,1,3},

  /* 3: marked dual-joining in isolated form */
  {0,0,0,8}, {1,0,1,8}, {4,1,1,3}, {2,1,0,4}, {5,0,1,3}, {3,0,0,8}, {7,1,1,3}, {6,0,1,3},

  /* 4: marked dual-joining in final form */
  {0,0,0,8}, {1,0,2,8}, {4,1,2,3}, {2,1,0,4}, {5,0,2,3}, {4,0,0,8}, {7,1,2,3}, {6,0,2,3},

  /* 5: seen right-joining */
  {0,0,0,8}, {1,0,0,8}, {3,1,0,4}, {2,1,0,4}, {5,0,0,4}, {5,0,0,8}, {7,1,0,6}, {6,0,0,4},

  /* 6: seen dalath/rish */
  {0,0,0,8}, {1,0,0,8}, {3,1,0,4}, {2,1,0,4}, {5,0,0,4}, {6,0,0,8}, {7,1,0,7}, {6,0,0,4},

  /* 7: marked alaph in final form */
  {0,0,0,8}, {1,0,5,8}, {3,1,5,4}, {2,1,5,4}, {5,0,5,8}, {7,0,0,8}, {5,0,5,8}, {6,0,5,8}
};

/*
 * To avoid massaging the feature list or setting glyph properties in normal cases,
 * we put the RTL-only features at the start of the array, the LTR ones at the end,
 * set the count to include only one kind but not the other, and adjust the base
 * if direction turns out to be LTR.
 */
static feature_setting_t std_arabic_features[] = {
  { HB_TAG('r','t','l','m'), 1 }, /* these two RTL features must be first, see above */
  { HB_TAG('r','t','l','a'), 1 },
  { HB_TAG('i','n','i','t'), INIT_FEATURE_BIT },
  { HB_TAG('m','e','d','i'), MEDI_FEATURE_BIT },
  { HB_TAG('f','i','n','a'), FINA_FEATURE_BIT },
  { HB_TAG('i','s','o','l'), ISOL_FEATURE_BIT },
  { HB_TAG('m','e','d','2'), MED2_FEATURE_BIT },
  { HB_TAG('f','i','n','2'), FIN2_FEATURE_BIT },
  { HB_TAG('f','i','n','3'), FIN3_FEATURE_BIT },
  { HB_TAG('c','c','m','p'), 1 },
  { HB_TAG('l','o','c','l'), 1 },
  { HB_TAG('r','l','i','g'), 1 },
  { HB_TAG('l','i','g','a'), 1 },
  { HB_TAG('c','a','l','t'), 1 },
  { HB_TAG('c','l','i','g'), 1 },
  { HB_TAG('c','s','w','h'), 1 },
  { HB_TAG('m','s','e','t'), 1 },
  { HB_TAG('c','u','r','s'), 1 },
  { HB_TAG('k','e','r','n'), 1 },
  { HB_TAG('m','a','r','k'), 1 },
  { HB_TAG('m','k','m','k'), 1 },
  { HB_TAG('l','t','r','m'), 1 }, /* these two LTR features must be last, see above */
  { HB_TAG('l','t','r','a'), 1 }
};
#define DIRECTIONAL_FEATURE_COUNT 2

void
hb_shape_arabic(hb_ot_layout_t * layout,
		hb_buffer_t * buffer,
		const char * cmap,
		const void * shaping_data,
		int direction,
		const feature_setting_t * user_features,
		unsigned int user_feature_count)
{
  const feature_setting_t * orig_features = &std_arabic_features[0];
  unsigned int feature_count = sizeof(std_arabic_features) / sizeof(std_arabic_features[0]) - DIRECTIONAL_FEATURE_COUNT;
  const feature_setting_t * features;
  hb_glyph_info_t * marked = NULL; /* will be set before use, unless state table logic is bad! */
  unsigned int state_row_offset = 0;
  unsigned int i;
  const char * cmap_subtable = hb_cmap_find_preferred_subtable(cmap, NULL);
  
  if (!cmap_subtable)
    return;
  
  if (direction > 0) {
    orig_features += DIRECTIONAL_FEATURE_COUNT;
  }
  features = hb_merge_user_features(orig_features, &feature_count, user_features, user_feature_count);

  for (i = 0; i < buffer->in_length; ++i) {
    hb_glyph_info_t * current = &buffer->in_string[i];
    hb_codepoint_t ch = current->gindex;
    const st_entry * entry;
    hb_codepoint_t glyph;
    
    if (direction < 0) {
      /* use mirrored char code unless it is missing from the font */
      hb_codepoint_t mirrored_char = hb_get_mirrored_char_code(ch);
      glyph = hb_cmap_subtable_char_to_glyph(cmap_subtable, mirrored_char);
      if (glyph) {
        /* found a mirrored form, so we'll use its join_type below */
        ch = mirrored_char;
      } else if (mirrored_char != ch) {
	/* fall back to original character code */
	glyph = hb_cmap_subtable_char_to_glyph(cmap_subtable, ch);
      }
    } else {
      glyph = hb_cmap_subtable_char_to_glyph(cmap_subtable, ch);
    }
    
    current->gindex = glyph;

    entry = &arabic_state_table[state_row_offset + hb_get_arabic_join_type(ch)];

    if (entry->curr_mask_index) {
      current->properties = mask_array[entry->curr_mask_index];
    }
    if (entry->prev_mask_index) {
      marked->properties = mask_array[entry->prev_mask_index];
    }
    if (entry->flags) {
      marked = &buffer->in_string[i];
    }

    state_row_offset = (unsigned int)entry->next_state * HB_NUM_JOIN_TYPES;
  }
  
  if (hb_ot_layout_has_substitution(layout)) {
    hb_apply_features(layout, HB_OT_LAYOUT_TABLE_TYPE_GSUB, NO_INDEX, NO_INDEX,
		   features, feature_count, buffer, &hb_ot_layout_substitute_lookup);
  }
  
  if (hb_ot_layout_has_positioning(layout)) {
    _hb_buffer_clear_positions(buffer);
    hb_apply_features(layout, HB_OT_LAYOUT_TABLE_TYPE_GPOS, NO_INDEX, NO_INDEX,
		   features, feature_count, buffer, &hb_ot_layout_position_lookup);
  }
  
  if (features != orig_features) {
    hb_free_feature_list(features);
  }
}
