/*
 * Copyright 2009 Mozilla Corporation.
 *
 * This is part of HarfBuzz, an OpenType Layout engine library.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Created by Jonathan Kew on 21/06/2009.
 */

#include "hb-shape-generic.h"
#include "hb-shaper-private.h"
#include "hb-cmap.h"
#include "hb-ompl.h"

/*
 * To avoid massaging the feature list or setting glyph properties in normal cases,
 * we put the LTR-only features at the start of the array, the RTL ones at the end,
 * and adjust the pointer and length to include one kind but not the other depending
 * on the directionality.
 */
static const feature_setting_t generic_features[] = {
  { HB_TAG('l','t','r','m'), 1 }, /* these two LTR features must be first, see above */
  { HB_TAG('l','t','r','a'), 1 },
  { HB_TAG('c','c','m','p'), 1 }, /* features used by default for either direction */
  { HB_TAG('l','o','c','l'), 1 },
  { HB_TAG('r','l','i','g'), 1 },
  { HB_TAG('l','i','g','a'), 1 },
  { HB_TAG('c','l','i','g'), 1 },
  { HB_TAG('c','a','l','t'), 1 },
  { HB_TAG('c','s','w','h'), 1 },
  { HB_TAG('c','u','r','s'), 1 },
  { HB_TAG('k','e','r','n'), 1 },
  { HB_TAG('m','a','r','k'), 1 },
  { HB_TAG('m','k','m','k'), 1 },
  { HB_TAG('r','t','l','m'), 1 }, /* these two RTL features must be last, see above */
  { HB_TAG('r','t','l','a'), 1 }
};
#define DIRECTIONAL_FEATURE_COUNT 2

void
hb_shape_generic(hb_ot_layout_t * layout,
		 hb_buffer_t * buffer,
		 const char * cmap,
		 const void * shaping_data,
		 int direction,
		 const feature_setting_t * user_features,
		 unsigned int user_feature_count)
{
  const feature_setting_t * orig_features = &generic_features[0];
  unsigned int feature_count = sizeof(generic_features) / sizeof(generic_features[0]) - DIRECTIONAL_FEATURE_COUNT;
  const feature_setting_t * features;
  unsigned int i;
  const char * cmap_subtable = hb_cmap_find_preferred_subtable(cmap, NULL);
  
  if (!cmap_subtable)
    return;
  
  if (direction < 0) {
    orig_features += DIRECTIONAL_FEATURE_COUNT;
  }
  features = hb_merge_user_features(orig_features, &feature_count, user_features, user_feature_count);
  
  /* test direction outside the loop rather than on every character */
  if (direction < 0) {
    for (i = 0; i < buffer->in_length; ++i) {
      /* use mirrored char code unless it is missing from the font */
      hb_codepoint_t ch = buffer->in_string[i].gindex;
      hb_codepoint_t mirrored_char = hb_get_mirrored_char_code(ch);
      hb_codepoint_t glyph = hb_cmap_subtable_char_to_glyph(cmap_subtable, mirrored_char);
      if (!glyph && mirrored_char != ch) {
	glyph = hb_cmap_subtable_char_to_glyph(cmap_subtable, ch);
      }
      buffer->in_string[i].gindex = glyph;
    }
  } else {
    for (i = 0; i < buffer->in_length; ++i) {
      buffer->in_string[i].gindex = hb_cmap_subtable_char_to_glyph(cmap_subtable, buffer->in_string[i].gindex);
    }
  }
  
  if (hb_ot_layout_has_substitution(layout)) {
    hb_apply_features(layout, HB_OT_LAYOUT_TABLE_TYPE_GSUB, NO_INDEX, NO_INDEX,
		   features, feature_count, buffer, &hb_ot_layout_substitute_lookup);
  }
  
  if (hb_ot_layout_has_positioning(layout)) {
    _hb_buffer_clear_positions(buffer);
    hb_apply_features(layout, HB_OT_LAYOUT_TABLE_TYPE_GPOS, NO_INDEX, NO_INDEX,
		   features, feature_count, buffer, &hb_ot_layout_position_lookup);
  }
  
  if (features != orig_features) {
    hb_free_feature_list(features);
  }
}
