/*
 * Copyright 2009 Mozilla Corporation.
 *
 * This is part of HarfBuzz, an OpenType Layout engine library.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Created by Jonathan Kew on 21/06/2009.
 */

//#define USE_INLINE_QSORT 1
#undef USE_INLINE_QSORT

#include "hb-shaper-private.h"

#ifdef USE_INLINE_QSORT
#include "qsort.h" /* fast inline qsort implementation */
#else
#include <stdlib.h>
#endif

#include <string.h> /* for memcpy */

struct lookup_record {
  unsigned int                index;
  hb_ot_layout_feature_mask_t mask;
};
typedef struct lookup_record lookup_record;
#define HB_INVALID_LOOKUP_INDEX 0xffffffffUL

#ifdef USE_INLINE_QSORT
#define lookup_LT(a,b) ((a)->index < (b)->index)
#else
int lookup_compare(const void *a, const void *b)
{
  return ((const lookup_record *)a)->index - ((const lookup_record *)b)->index;
}
#endif

/*
 * Find all the lookups required to implement a set of features, and apply to the buffer.
 */
void
hb_apply_features(hb_ot_layout_t *layout,
		  hb_ot_layout_table_type_t table_type,
		  unsigned int script_index,
		  unsigned int language_index,
		  const feature_setting_t *features,
		  int num_features,
		  hb_buffer_t *buffer,
		  hb_lookup_applicator_func_t apply_function)
{
#define AUTO_LOOKUP_COUNT 128 /* probably enough for most real-world fonts */
#define ALLOC_EXTRA	  128 /* if we do need to allocate separately, grab this much extra space */
  
  lookup_record auto_lookups[AUTO_LOOKUP_COUNT];
  lookup_record * lookups = &auto_lookups[0];
  unsigned int lookup_alloc_count = AUTO_LOOKUP_COUNT;
  unsigned int num_lookups = 0;

  unsigned int feature_index;
  unsigned int lookup_count;
  unsigned int lookup_index;
  unsigned int i, j;
  
  /* get lookup(s) for the required feature, if any */
  if (hb_ot_layout_language_get_required_feature_index(layout, table_type,
						       script_index, language_index,
						       &feature_index)) {
    lookup_count = hb_ot_layout_feature_get_lookup_count(layout, table_type, feature_index);
    
    /*
     * Check that we have room in our lookup array, reallocate if necessary.
     * Highly unlikely except in the case of a corrupt or malicious font that claims to have a huge lookup_count.
     */
    if (lookup_count >= lookup_alloc_count) {
      lookup_record * new_lookups = malloc((lookup_count + ALLOC_EXTRA) * sizeof(lookups[0]));
      if (new_lookups) {
	memcpy(new_lookups, lookups, lookup_alloc_count * sizeof(lookups[0]));
	lookup_alloc_count = lookup_count + ALLOC_EXTRA;
	/* no need to free the old pointer, as it must be our auto-array at this point */
	lookups = new_lookups;
      } else {
	/* not enough memory, just truncate the list to avoid overstepping the array */
	lookup_count = lookup_alloc_count - 1;
      }
    }
    
    for (j = 0; j < lookup_count; ++j) {
      lookup_index = hb_ot_layout_feature_get_lookup_index(layout, table_type, feature_index, j);
      lookups[num_lookups].index = lookup_index;
      lookups[num_lookups].mask = ~0;
      ++num_lookups;
    }
  }
  
  /* find lookups for the requested features */
  for (i = 0; i < num_features; ++i) {
    if (!features[i].value) {
      continue; /* disabled feature entry */
    }
    if ((script_index == NO_INDEX && 
	 hb_ot_layout_table_find_feature(layout, table_type,
					 features[i].tag, &feature_index)) ||
	hb_ot_layout_language_find_feature(layout, table_type,
					   script_index, language_index,
					   features[i].tag, &feature_index))
    {
      lookup_count = hb_ot_layout_feature_get_lookup_count(layout, table_type, feature_index);
      
      /* check that we have room in our lookup array, reallocate if necessary */
      if (num_lookups + lookup_count >= lookup_alloc_count) {
	lookup_record * new_lookups = malloc((num_lookups + lookup_count + ALLOC_EXTRA) * sizeof(lookups[0]));
	if (!new_lookups) {
	  break; /* failed, just give up collecting lookups */
	}
	memcpy(new_lookups, lookups, lookup_alloc_count * sizeof(lookups[0]));
	lookup_alloc_count = num_lookups + lookup_count + ALLOC_EXTRA;
	if (lookups != &auto_lookups[0]) {
	  free(lookups);
	}
	lookups = new_lookups;
      }
      
      /* store the lookups into the array, now that we know there's room */
      for (j = 0; j < lookup_count; ++j) {
	lookup_index = hb_ot_layout_feature_get_lookup_index(layout, table_type, feature_index, j);
	lookups[num_lookups].index = lookup_index;
	lookups[num_lookups].mask = features[i].value;
	++num_lookups;
      }
    }
  }
  
  /* sort lookups into correct order to apply */
#ifdef USE_INLINE_QSORT
  QSORT(lookup_record, &lookups[0], num_lookups, lookup_LT);
#else
  qsort(&lookups[0], num_lookups, sizeof(lookups[0]), lookup_compare);
#endif
  
  /* eliminate any duplicates by changing their index to HB_INVALID_LOOKUP_INDEX */
  for (i = 1; i < num_lookups; ++i) {
    if (lookups[i-1].index == lookups[i].index) {
      lookups[i].index = HB_INVALID_LOOKUP_INDEX;
      lookups[i-1].mask &= lookups[i].mask;
    }
  }
  
  /* finally, apply the lookups across the buffer */
  for (i = 0; i < num_lookups; ++i) {
    if (lookups[i].index != HB_INVALID_LOOKUP_INDEX) {
      (*apply_function)(layout, buffer, lookups[i].index, lookups[i].mask);
    }
  }
  
  if (lookups != &auto_lookups[0]) {
    free(lookups);
  }
}

const feature_setting_t *
hb_merge_user_features(const feature_setting_t * orig_features, unsigned int * feature_count,
		    const feature_setting_t * user_features, unsigned int user_feature_count)
{
  unsigned int i, j;
  feature_setting_t *merged_features;
  if (!user_features || user_feature_count == 0) {
    return orig_features;
  }
  
  merged_features = (feature_setting_t*)malloc((*feature_count + user_feature_count) * sizeof(feature_setting_t));
  if (!merged_features) {
    /* either memory is so tight we're probably doomed, or else user_feature_count was unreasonable */
    return orig_features;
  }

  memcpy(merged_features, orig_features, *feature_count * sizeof(feature_setting_t));
  
  /*
   * For each new feature, add it to the list, or update the existing value if it's already present.
   * Note that features with value == 0 will in effect be disabled. We don't bother removing them
   * from the list because there probably won't be many, and this saves us shifting the rest of the
   * entries. We'll just skip them later when collecting the list of lookups.
   */
  for (i = 0; i < user_feature_count; ++i) {
    for (j = 0; j < *feature_count; ++j) {
      if (merged_features[j].tag == user_features[i].tag) {
	merged_features[j].value = user_features[i].value;
	break;
      }
    }
    if (j == *feature_count) {
      merged_features[j].tag = user_features[i].tag;
      merged_features[j].value = user_features[i].value;
      (*feature_count)++;
    }
  }
  
  return merged_features;
}

void
hb_free_feature_list(const feature_setting_t * features)
{
  free((void*)features);
}
